#!/usr/bin/python
#
# Random 2d-projections of streaming point data from a file
#
# Usage: projectPoints file [leading dimensions to skip]
#
# Author: Michael Robinson
#
# 2011-03-30: Initial version
#
# Copyright (c) 2011, Michael Robinson
# Distribution of unaltered copies permitted for noncommercial use only
# All other uses require express permission of the author
# This software comes with no warrantees express or implied 

import gtk
try:
    import hildon
except ImportError:
    hildon=None
import sys
import threading
import time
import gobject
import numpy
import numpy.random
from numpy import sqrt

# Determine how to set scale bounds for track
def scale_track(track,width,height):
    xs=[x[0] for x in track]
    ys=[x[1] for x in track]
    cx=numpy.mean(xs)
    cy=numpy.mean(ys)
    sx=numpy.std(xs)
    sy=numpy.std(ys)
    maxx=cx+2*sx
    minx=cx-2*sx
    maxy=cy+2*sy
    miny=cy-2*sy

    return (width/(maxx-minx),height/(maxy-miny),minx,miny)

# Random projection matrix
def randproj(dims):
    p1=numpy.random.randn(dims,1)
    p2=numpy.random.randn(dims,1)

    p1=p1/numpy.sqrt(numpy.sum(p1**2))
    p2=p2-numpy.sum(p1*p2)*p1
    p2=p2/numpy.sqrt(numpy.sum(p2**2))
    
    return numpy.concatenate((p1,p2),1)

# Thread containing block of code that reads from a file and then follows it
class readFollower(threading.Thread):
    stopthread = threading.Event()

    def __init__(self,filepointer,callback):
        threading.Thread.__init__(self) # Reuse thread constructor
        self.fp=filepointer
        self.cb=callback

    def run(self):
        for line in self.tail_f():         # Look at the file
            self.cb(line)  # Invoke callback to send back data we read

    def stop(self): # Kill the thread on request
        self.stopthread.set()

    def tail_f(self):  # This reads the file, and then follows it
        interval = 0.1 # Seconds between updates once following
        for line in self.fp:
            yield line
        while not self.stopthread.isSet(): # Exit on request
            line = self.fp.readline()
            if not line:
                time.sleep(interval)
            else:
                yield line

# Display window
class projectPointsWindow:
    def delete_event(self, event, data=None):
        return False

    def main(self):
        gtk.main()
        return 0

    def destroy_event(self, data=None):
        self.reader.stop()
        gtk.main_quit()

    def add_line(self,line):  # Add a new line of input data
        if self.listening:
            try:
              listy=map(float,line.split(','))[self.skipdims:]
            except ValueError:
              print "Skipping nonnumeric data"
              return False
              
            parsed=numpy.array([listy])
            currows=numpy.size(self.dataBlock,0)
            if currows > 0:
              self.dataBlock=numpy.concatenate((self.dataBlock,parsed))
            else:
              self.dims=numpy.size(parsed,1)
              self.dataBlock=parsed
              
              # Random projection
              self.proj=randproj(self.dims)

            gobject.idle_add(self.update) # Update the display with this new data
        return True

    def update(self): # Update the display
        # Color state
        style=self.screen.get_style()
        gc=style.fg_gc[gtk.STATE_NORMAL]
        gc_oldfg = gc.foreground

        # Erase current display
        gc.foreground=self.black
        self.pixmap.draw_rectangle(gc,True,0,0,512,380)

        # Use projection to compress
        datap=numpy.dot(self.dataBlock,self.proj)
        track=[(datap[i,0],datap[i,1]) for i in range(0,datap.shape[0])]

        # Window viewport, based on data
        (sx,sy,xm,ym)=scale_track(track,512,380)

        # Plot points
        gc.foreground=self.white
        #self.pixmap.draw_points(gc,[(int((x[0]-xm)*sx),int((x[1]-ym)*sy)) for x in track])
        for pt in track:
            x=int((((pt[0]-xm)*sx)-256)*self.zoom+256)
            y=int((((pt[1]-ym)*sy)-160)*self.zoom+160)
            self.pixmap.draw_rectangle(gc,True,x-2,y-2,5,5)

        # Most recent location
        if len(track) > 0:
            gc.foreground=self.red
            newpt=track[-1]
            x=int((((newpt[0]-xm)*sx)-256)*self.zoom+256)
            y=int((((newpt[1]-ym)*sy)-160)*self.zoom+160)
            self.pixmap.draw_rectangle(gc,True,x-5,y-5,11,11)

        # Restore foreground
        gc.foreground=gc_oldfg

        # Update the display
        self.screen.window.draw_drawable(self.screen.get_style().fg_gc[gtk.STATE_NORMAL], self.pixmap, 0, 0, 0, 0, 512, 380)
        return False

    def startstop_event(self,event): # Toggle accepting new data
        if self.listening:
            self.start_button.set_label('Start')
            self.listening = False
        else:
            self.start_button.set_label('Stop')
            self.listening = True
        gobject.idle_add(self.update)
        return True

    def clear_event(self,event): # Clear the data queue
        self.dataBlock=numpy.zeros((1,self.dims))
        gobject.idle_add(self.update)
        return True

    def newproj_event(self,event): # New random projection
        self.proj=randproj(self.dims)
        gobject.idle_add(self.update)
        return True
        
    def in_event(self,event):
        self.zoom=self.zoom*1.1
        return True
    
    def out_event(self,event):
        self.zoom=self.zoom/1.1
        return True

    def expose_event(self,event,data):
        gobject.idle_add(self.update)
        return False

    # Constructor requests input datafile and number of dimensions in data
    def __init__(self,file=sys.stdin,skipdims=0,dims=4):  
        if hildon:  # Construct the window, depending on how we're running
            self.window = hildon.Window()
        else:
            self.window = gtk.Window()

        # Start out by listening to the datafile
        self.listening=True

        # Initial data area
        self.dims=dims-skipdims
        self.skipdims=skipdims
        self.dataBlock=numpy.zeros((0,self.dims))
        
        # Random projection
        self.proj=randproj(self.dims)
        self.zoom=1

        # Window boilerplate
        self.window.set_title("Random projection of sampled data")
        self.window.connect("delete_event",self.delete_event)
        self.window.connect("destroy",self.destroy_event)
        self.window.connect("expose_event",self.expose_event)
        self.window.set_border_width(5)

        # Display area boilerplate
        self.screen=gtk.DrawingArea()
        self.screen.set_size_request(512,380)

        # Widgets on the display
        hbox=gtk.HBox()
        vbox=gtk.VBox()
        hbox.pack_start(self.screen,True,True,0)
        hbox.pack_start(vbox,True,True,0)
        clear_button=gtk.Button('Clear')
        clear_button.connect('clicked',self.clear_event)
        vbox.pack_start(clear_button,True,True,0)
        self.start_button=gtk.Button('Stop')
        self.start_button.connect('clicked',self.startstop_event)
        vbox.pack_start(self.start_button,True,True,0)
        proj_button=gtk.Button('New Proj')
        proj_button.connect('clicked',self.newproj_event)
        vbox.pack_start(proj_button,True,True,0)
        in_button=gtk.Button('+')
        in_button.connect('clicked',self.in_event)
        out_button=gtk.Button('-')
        out_button.connect('clicked',self.out_event)
        zoom_box=gtk.HBox()
        zoom_box.pack_start(in_button,True,True,0)
        zoom_box.pack_start(out_button,True,True,0)
        vbox.pack_start(zoom_box,True,True,0)

        self.window.add(hbox)
        self.window.show_all()

        # Backing pixmap
        self.pixmap = gtk.gdk.Pixmap(self.screen.window, 512, 380)
        self.pixmap.draw_rectangle(self.screen.get_style().black_gc,True,0,0,512,380)

        # Useful colors for drawing
        self.black=self.screen.window.get_colormap().alloc_color(0,0,0)
        self.white=self.screen.window.get_colormap().alloc_color(0xFFFF,0xFFFF,0xFFFF)
        self.red=self.screen.window.get_colormap().alloc_color(0xFFFF,0,0)
        self.green=self.screen.window.get_colormap().alloc_color(0,0xFFFF,0)
        self.blue=self.screen.window.get_colormap().alloc_color(0,0,0xFFFF)

        # Start up the reader
        self.reader=readFollower(file,self.add_line)
        self.reader.start()

# Main function
if __name__ == "__main__":
    gtk.gdk.threads_init()
    if len(sys.argv) < 3 or sys.argv[2] == '':
        ppw=projectPointsWindow(open(sys.argv[1]))
    else:
        ppw=projectPointsWindow(open(sys.argv[1]),int(sys.argv[2]))
    ppw.main()
