#!/usr/bin/python
# Process a sequence of coded pulses in real time
# making the assumption that the pulses come in blocks 
# of a fixed length
#
# Command line:
# rx_tdma_chirp [-m mode] [-a] [-c | -s] [-o output] [-t transmitters]
#
# Chirp reference files are hard coded
#
# Copyright (c) 2011, Michael Robinson
# Distribution of unaltered copies permitted for noncommercial use only
# All other uses require express permission of the author
# This software comes with no warrantees express or implied 
#
# Version 0.1

import threading
import gobject
import pygst
pygst.require("0.10")
import pango
import gst
import time
import struct
import wave
import numpy
import os, sys, getopt
from math import sqrt
from numpy import conj
from numpy.fft import fft, ifft
from scipy.signal import fftconvolve

# Unpack data from gstreamer
def unpack(str,bytesPerSample=2):
    result=list()
    for i in range(0,len(str)-1,bytesPerSample):
        result.append(struct.unpack('h',str[i:i+bytesPerSample])[0])
    return result

class chirpTDMAProcessor:
    def __init__(self,numberTX,chirpRefFile,mode='both',outputFile=None,syncRefFile=None,correlate=False,normalize=True):
        # Parameters from user
        self.numberTX=numberTX
        self.outputFile=outputFile
        self.mode=mode
        self.file=outputFile
        self.correlate=correlate
        self.normalize=normalize

        # Transform parameters
        self.txPeriod=6450
        self.searchWindow=int(self.txPeriod/3)
        self.blockSize=self.txPeriod*self.numberTX
        self.sampleRate=44100
        self.dataBlock=numpy.zeros((self.blockSize))

        # Load reference signal from file
        if chirpRefFile==None:
            self.ref=None
        else:
            f=wave.open(chirpRefFile,'r')

            # Unpack file into a numpy array
            wavdata=unpack(f.readframes(f.getnframes()),f.getsampwidth())
            self.ref=numpy.array(wavdata) #numpy.conjugate(fft(wavdata,self.blockSize))
            f.close()

        # If there is a special synchronization pulse
        if syncRefFile == None:
            self.syncRef = None
        else:
            f=wave.open(syncRefFile,'r')
            # Unpack file into a numpy array
            wavdata=unpack(f.readframes(f.getnframes()),f.getsampwidth())
            self.syncRef=numpy.array(wavdata)
            f.close()

        # Construct gstreamer pipeline to funnel data into the application
        # pulsesrc ! audioconvert ! fakesink ! (this program)
        self.pipeline=gst.Pipeline("mypipeline")

        src=gst.element_factory_make("pulsesrc", "src")
        self.pipeline.add(src)

        ac=gst.element_factory_make("audioconvert","ac")
        self.pipeline.add(ac)

        sink=gst.element_factory_make("fakesink","fakesink")
        self.pipeline.add(sink)

        src.link(ac,gst.caps_from_string("audio/x-raw-int,width=16,signed=true,rate="+ str(self.sampleRate) + ",channels=1"))
        ac.link(sink)
        src.set_property("blocksize",self.blockSize*2)
        
        sink.set_property("signal-handoffs",True)
        sink.connect("handoff",self.buffer_cb)
        self.ready=threading.Event()
        self.ready.set()
        return

    def run(self):
        # Activate Gstreamer pipeline
        self.pipeline.set_state(gst.STATE_PLAYING)
        while True:
            try:
                time.sleep(0.1)
            except (KeyboardInterrupt, SystemExit):
                self.ready.wait()
                self.pipeline.set_state(gst.STATE_NULL)
                break
        return True

    def buffer_cb(self, buffer, pad, user_data):
        self.ready.clear()
        # Unpack and store data
        data=numpy.array([x*190.0/2.0**16 for x in unpack(pad)])

        # Correlate against stored reference if desired
        if self.correlate:
            data_cor=abs(fftconvolve(data,self.ref
            
            ,mode='same'))
        else:
            data_cor=abs(data) # No correlation (better for certain waveforms)

        # Peak detect to locate one of the pulses
        if self.syncRef == None:
            peakDelay=numpy.argmax(data_cor) # Using the same matched filter tap as before
        else:
            # Correlate against sync reference. Peak is likely to be the synchronization chirp
            data_sync=abs(fftconvolve(data,self.syncRef,mode='same'))
            peakDelay=numpy.argmax(data_sync)

        # Convert to a relative delay
        peakRelDelay=numpy.mod(peakDelay,self.txPeriod)

        # Search for other peaks in the vicinity of plus or minus an integer number of periods
        outline=''
        times=[0]*self.numberTX
        levels=[0]*self.numberTX
        mxidx=0
        for i in range(0,self.numberTX):
            # Search window
            center=i*self.txPeriod+peakRelDelay
            wst=center-self.searchWindow
            wen=center+self.searchWindow

            # Boundary conditions
            if wst<0: wst=0
            if wen>=data_cor.size: wen=data_cor.size-1

            # Search for peak
            idx=numpy.argmax(data_cor[wst:wen])+wst

            # Features of peak: timing and level
            times[i]=idx-center
            levels[i]=10*numpy.log10(numpy.max(abs(data_cor[idx])))+10
            
            # Maximum level (useful if we need to normalize signals
            if levels[i]>levels[mxidx]: mxidx=i

        # Normalize signals if desired
        if self.normalize:
            levels=[x-levels[mxidx] for x in levels]

        # Forget slow-time synchronization, which seems to go bad due to loss of 
        # synchronization from gstreamer's handoff.  This results in the maximum signal 
        # received being in the first slot
        dat=numpy.roll(numpy.column_stack((numpy.array(times),numpy.array(levels))),-mxidx,axis=0)
        times=dat[:,0]
        levels=dat[:,1]

        # Produce string, according to user desires
        for i in range(0,self.numberTX):
            if i>0:
                outline += ','
            if self.mode=='both' or self.mode == 'timing': # Timing
                outline += str(times[i])
            if self.mode=='both' or self.mode == 'level': # Signal level
                outline += str(int(levels[i]*10.0)/10.0)

        # Generate the output to screen and to file
        # Use system call rather than direct write to file as this permits "tail -f" 
        # to keep track of file changes easier, and avoids problems with closing 
        # the file if this program crashes 
        print outline
        os.system('echo ' + outline + ' >>' + self.file) 
        self.ready.set()

        return True

# Main function
if __name__ == "__main__":
    gobject.threads_init()

    try:
        opts,args = getopt.getopt(sys.argv[1:],"m:csao:t:")
    except:
        print "Usage: rx_tdma_chirp [-m mode] [-c | -s] [-a] [-o output] [-t transmitters]\n\n\
-m : mode is one of the following: both (default), timing, or level\n\
-a : do not normalize to maximum received signal\n\
-c : Correlate against chirp type 1 (default uses no matched filtering)\n\
-s : Uses chirp type 2 for synchronization (implies -c)\n\
-t : Sets number of transmitters (5 default)\n\
-o : Sets output file (defaults to out.txt)\n\n\
Matched filters use hard coded paths to stored chirp reference files\n"
        sys.exit(2)

    mode='both'
    outputFile='out.txt'
    sync=False
    numTx=5
    corr=False
    norm=True

    for o,a in opts:
        if o == '-m':
            mode=a
        elif o == '-s':
            corr=True
            sync=True
        elif o == '-o':
            outputFile=a
        elif o == '-t':
            numTx=int(a)
        elif o == '-a':
            norm = False
        elif o == '-c':
            corr=True

    if sync and corr:
        ctp=chirpTDMAProcessor(numTx,'../../data/chirp_transmitters/ref1b.wav',mode,
                               outputFile,syncRefFile='../../data/chirp_transmitters/ref2b.wav',
                               correlate=corr,normalize=norm)
    elif sync:
        ctp=chirpTDMAProcessor(numTx,'../../data/chirp_transmitters/ref1b.wav',mode,outputFile,
                               correlate=corr,normalize=norm)
    else:
        ctp=chirpTDMAProcessor(numTx,None,mode,
                               outputFile,syncRefFile=None,
                               correlate=corr,normalize=norm)
        
    ctp.run()
